from flask import Flask, jsonify, request
from requests.exceptions import ReadTimeout

import mailadm.db
from mailadm.conn import DBError
from mailadm.mailcow import MailcowError


def create_app_from_db_path(db_path=None):
    if db_path is None:
        db_path = mailadm.db.get_db_path()

    db = mailadm.db.DB(db_path)
    return create_app_from_db(db)


def create_app_from_db(db):
    app = Flask("mailadm-account-server")
    app.db = db

    @app.route("/", methods=["POST"])
    def new_email():
        token = request.args.get("t")
        if token is None:
            return (
                jsonify(type="error", status_code=403, reason="?t (token) parameter not specified"),
                403,
            )

        with db.write_transaction() as conn:
            token_info = conn.get_tokeninfo_by_token(token)
            if token_info is None:
                return (
                    jsonify(
                        type="error",
                        status_code=403,
                        reason="token {} is invalid".format(token),
                    ),
                    403,
                )
            try:
                user_info = conn.add_email_account_tries(token_info, tries=10)
                return jsonify(
                    email=user_info.addr,
                    password=user_info.password,
                    expiry=token_info.expiry,
                    ttl=user_info.ttl,
                )
            except (DBError, MailcowError) as e:
                if "does already exist" in str(e):
                    return (
                        jsonify(
                            type="error",
                            status_code=409,
                            reason="user already exists in mailcow",
                        ),
                        409,
                    )
                if "UNIQUE constraint failed" in str(e):
                    return (
                        jsonify(
                            type="error",
                            status_code=409,
                            reason="user already exists in mailadm",
                        ),
                        409,
                    )
                return jsonify(type="error", status_code=500, reason=str(e)), 500
            except ReadTimeout:
                return jsonify(type="error", status_code=504, reason="mailcow not reachable"), 504

    return app



import time

from mailadm.conn import DBError
from mailadm.gen_qr import gen_qr
from mailadm.mailcow import MailcowError
from mailadm.util import get_human_readable_id


def add_token(db, name, expiry, maxuse, prefix, token) -> dict:
    """Adds a token to create users"""
    if token is None:
        token = expiry + "_" + get_human_readable_id(len=15)
    with db.write_transaction() as conn:
        try:
            info = conn.add_token(
                name=name,
                token=token,
                expiry=expiry,
                maxuse=maxuse,
                prefix=prefix,
            )
        except DBError as e:
            return {"status": "error", "message": "failed to add token {}: {}".format(name, e)}
        except ValueError:
            return {"status": "error", "message": "maxuse must be a number"}
        tc = conn.get_tokeninfo_by_name(info.name)
        return {"status": "success", "message": dump_token_info(tc)}


def add_user(db, token=None, addr=None, password=None, dryrun=False) -> {}:
    """Adds a new user to be managed by mailadm"""
    with db.write_transaction() as conn:
        if token is None:
            if "@" not in addr:
                # there is probably a more pythonic solution to this.
                # the goal is to display the error, whether the command came via CLI or delta bot.
                return {"status": "error", "message": "invalid email address: {}".format(addr)}

            token_info = conn.get_tokeninfo_by_addr(addr)
            if token_info is None:
                return {
                    "status": "error",
                    "message": "could not determine token for addr: {!r}".format(addr),
                }
        else:
            token_info = conn.get_tokeninfo_by_name(token)
            if token_info is None:
                return {"status": "error", "message": "token does not exist: {!r}".format(token)}
        try:
            user_info = conn.add_email_account(token_info, addr=addr, password=password)
        except DBError as e:
            return {
                "status": "error",
                "message": "failed to add e-mail account {}: {}".format(addr, e),
            }
        except MailcowError as e:
            return {
                "status": "error",
                "message": "failed to add e-mail account {}: {}".format(addr, e),
            }
        if dryrun:
            conn.delete_email_account(user_info.addr)
            return {"status": "dryrun", "message": user_info}
        return {"status": "success", "message": user_info}


def prune(db, dryrun=False) -> {}:
    sysdate = int(time.time())
    with db.read_connection() as conn:
        expired_users = conn.get_expired_users(sysdate)
        if not expired_users:
            return {"status": "success", "message": ["nothing to prune"]}
    if dryrun:
        result = {"status": "dryrun", "message": []}
        for user_info in expired_users:
            result["message"].append(
                "would delete %s (token %s)" % (user_info.addr, user_info.token_name),
            )
    else:
        result = {"status": "success", "message": []}
        for user_info in expired_users:
            try:
                with db.read_connection() as conn:
                    conn.get_mailcow_connection().del_user_mailcow(user_info.addr)
                with db.write_transaction() as conn:
                    conn.del_user_db(user_info.addr)
            except (DBError, MailcowError) as e:
                result["status"] = "error"
                result["message"].append("failed to delete account %s: %s" % (user_info.addr, e))
                continue
            result["message"].append(
                "pruned %s (token %s)" % (user_info.addr, user_info.token_name),
            )
    return result


def list_tokens(db) -> str:
    """Print token info for all tokens"""
    output = ["Existing tokens:\n"]
    with db.read_connection() as conn:
        for name in conn.get_token_list():
            token_info = conn.get_tokeninfo_by_name(name)
            output.append(dump_token_info(token_info))
    return "\n".join(output)


def qr_from_token(db, tokenname):
    with db.read_connection() as conn:
        token_info = conn.get_tokeninfo_by_name(tokenname)
        config = conn.config

    if token_info is None:
        return {"status": "error", "message": "token {!r} does not exist".format(tokenname)}

    image = gen_qr(config, token_info)
    fn = "docker-data/dcaccount-%s-%s.png" % (config.mail_domain, token_info.name)
    image.save(fn)
    return {"status": "success", "filename": fn}


def dump_token_info(token_info) -> str:
    """Format token info into a string"""
    return """token: {}
  address prefix: {}
  accounts expire after: {}
  token was used {} of {} times
  token: {}
    - url: {}
    - QR data: {}
    """.format(
        token_info.name,
        token_info.prefix,
        token_info.expiry,
        token_info.usecount,
        token_info.maxuse,
        token_info.token,
        token_info.get_web_url(),
        token_info.get_qr_uri(),
    )



import contextlib
import logging
import os
import sqlite3
import time
from pathlib import Path

from .conn import Connection


def get_db_path():
    db_path = os.environ.get("MAILADM_DB", "/mailadm/docker-data/mailadm.db")
    try:
        sqlite3.connect(db_path)
    except sqlite3.OperationalError:
        raise RuntimeError("mailadm.db not found: MAILADM_DB not set")
    return Path(db_path)


class DB:
    def __init__(self, path, autoinit=True, debug=False):
        self.path = path
        self.debug = debug
        self.ensure_tables()

    def _get_connection(self, write=False, transaction=False, closing=False):
        # we let the database serialize all writers at connection time
        # to play it very safe (we don't have massive amounts of writes).
        mode = "ro"
        if write:
            mode = "rw"
        if not self.path.exists():
            mode = "rwc"
        uri = "file:%s?mode=%s" % (self.path, mode)
        sqlconn = sqlite3.connect(
            uri,
            timeout=60,
            isolation_level=None if transaction else "DEFERRED",
            uri=True,
        )
        if self.debug:
            sqlconn.set_trace_callback(print)

        # Enable Write-Ahead Logging to avoid readers blocking writers and vice versa.
        if write:
            sqlconn.execute("PRAGMA journal_mode=wal")

        if transaction:
            start_time = time.time()
            while 1:
                try:
                    sqlconn.execute("begin immediate")
                    break
                except sqlite3.OperationalError:
                    # another thread may be writing, give it a chance to finish
                    time.sleep(0.1)
                    if time.time() - start_time > 5:
                        # if it takes this long, something is wrong
                        raise
        conn = Connection(sqlconn, self.path, write=write)
        if closing:
            conn = contextlib.closing(conn)
        return conn

    @contextlib.contextmanager
    def write_transaction(self):
        conn = self._get_connection(closing=False, write=True, transaction=True)
        try:
            yield conn
        except Exception:
            conn.rollback()
            conn.close()
            raise
        else:
            conn.commit()
            conn.close()

    def write_connection(self, closing=True):
        return self._get_connection(closing=closing, write=True)

    def read_connection(self, closing=True):
        return self._get_connection(closing=closing, write=False)

    def init_config(self, mail_domain, web_endpoint, mailcow_endpoint, mailcow_token):
        with self.write_transaction() as conn:
            conn.set_config("mail_domain", mail_domain)
            conn.set_config("web_endpoint", web_endpoint)
            conn.set_config("mailcow_endpoint", mailcow_endpoint)
            conn.set_config("mailcow_token", mailcow_token)

    def is_initialized(self):
        with self.read_connection() as conn:
            return conn.is_initialized()

    def get_config(self):
        with self.read_connection() as conn:
            return conn.config

    CURRENT_DBVERSION = 1

    def ensure_tables(self):
        with self.read_connection() as conn:
            if conn.get_dbversion():
                return
        with self.write_transaction() as conn:
            logging.info("DB: Creating tables %s", self.path)

            conn.execute(
                """
                CREATE TABLE tokens (
                    name TEXT PRIMARY KEY,
                    token TEXT NOT NULL UNIQUE,
                    expiry TEXT NOT NULL,
                    prefix TEXT,
                    maxuse INTEGER default 50,
                    usecount INTEGER default 0
                )
            """,
            )
            conn.execute(
                """
                CREATE TABLE users (
                    addr TEXT PRIMARY KEY,
                    date INTEGER,
                    ttl INTEGER,
                    token_name TEXT NOT NULL,
                    FOREIGN KEY (token_name) REFERENCES tokens (name)
                )
            """,
            )
            conn.execute(
                """
                CREATE TABLE config (
                    name TEXT PRIMARY KEY,
                    value TEXT
                )
            """,
            )
            conn.set_config("dbversion", self.CURRENT_DBVERSION)



import collections
import grp
import os
import pwd
import time
from pathlib import Path
from random import randint

import deltachat
import mailadm.bot
import mailadm.db
import pytest
from _pytest.pytester import LineMatcher


@pytest.fixture(autouse=True)
def _nocfg(monkeypatch):
    # tests can still set this env var but we want to isolate tests by default
    monkeypatch.delenv("MAILADM_DB", raising=False)

    def getpwnam(name):
        raise KeyError(name)

    monkeypatch.setattr(pwd, "getpwnam", getpwnam)


class ClickRunner:
    def __init__(self, main):
        from click.testing import CliRunner

        self.runner = CliRunner()
        self._main = main
        self._rootargs = []

    def set_basedir(self, account_dir):
        self._rootargs.insert(0, "--basedir")
        self._rootargs.insert(1, account_dir)

    def run_ok(self, args, fnl=None, input=None):
        __tracebackhide__ = True
        argv = self._rootargs + args
        # we use our nextbackup helper to cache account creation
        # unless --no-test-cache is specified
        res = self.runner.invoke(self._main, argv, catch_exceptions=False, input=input)
        if res.exit_code != 0:
            print(res.output)
            raise Exception("cmd exited with %d: %s" % (res.exit_code, argv))
        return _perform_match(res.output, fnl)

    def run_fail(self, args, fnl=None, input=None, code=None):
        __tracebackhide__ = True
        argv = self._rootargs + args
        res = self.runner.invoke(self._main, argv, catch_exceptions=False, input=input)
        if res.exit_code == 0 or (code is not None and res.exit_code != code):
            print(res.output)
            raise Exception(
                "got exit code {!r}, expected {!r}, output: {}".format(
                    res.exit_code,
                    code,
                    res.output,
                ),
            )
        return _perform_match(res.output, fnl)


def _perform_match(output, fnl):
    __tracebackhide__ = True
    if fnl:
        lm = LineMatcher(output.splitlines())
        lines = [x.strip() for x in fnl.strip().splitlines()]
        try:
            lm.fnmatch_lines(lines)
        except Exception:
            print(output)
            raise
    return output


@pytest.fixture
def cmd():
    """invoke a command line subcommand."""
    from mailadm.cmdline import mailadm_main

    return ClickRunner(mailadm_main)


@pytest.fixture
def db(tmpdir, make_db):
    path = tmpdir.ensure("base", dir=1)
    return make_db(path)


def prepare_account(addr, mailcow, db_path):
    password = mailcow.auth["X-API-Key"]
    mailcow.add_user_mailcow(addr, password, "admbot")
    ac = deltachat.Account(str(db_path))
    ac.run_account(addr, password)
    return ac


@pytest.fixture()
def admingroup(admbot, botadmin, db):
    admchat = admbot.create_group_chat("admins", [], verified=True)
    with db.write_transaction() as conn:
        conn.set_config("admingrpid", admchat.id)
    botplugin = mailadm.bot.AdmBot(db, admbot)
    admbot.add_account_plugin(botplugin)
    qr = admchat.get_join_qr()
    chat = botadmin.qr_join_chat(qr)
    while "added by" not in chat.get_messages()[len(chat.get_messages()) - 1].text:
        print(chat.get_messages()[len(chat.get_messages()) - 1].text)
        time.sleep(1)
    chat.admbot = admbot
    chat.botadmin = botadmin
    chat.botplugin = botplugin
    return chat


@pytest.fixture()
def admbot(mailcow, db, tmpdir, mailcow_domain):
    addr = "pytest-admbot-%s@%s" % (randint(0, 99999), mailcow_domain)
    tmpdir = Path(str(tmpdir))
    admbot_db_path = str(mailadm.bot.get_admbot_db_path(db_path=tmpdir.joinpath("admbot.db")))
    ac = prepare_account(addr, mailcow, admbot_db_path)
    ac._evlogger = ac.add_account_plugin(deltachat.events.FFIEventLogger(ac))
    ac.run_account(show_ffi=True)
    yield ac
    ac.shutdown()
    ac.wait_shutdown()
    mailcow.del_user_mailcow(addr)


@pytest.fixture
def botadmin(mailcow, db, tmpdir, mailcow_domain):
    addr = "pytest-admin-%s@%s" % (randint(0, 99999), mailcow_domain)
    tmpdir = Path(str(tmpdir))
    db_path = mailadm.bot.get_admbot_db_path(tmpdir.joinpath("botadmin.db"))
    ac = prepare_account(addr, mailcow, db_path)
    yield ac
    ac.shutdown()
    ac.wait_shutdown()
    mailcow.del_user_mailcow(addr)


@pytest.fixture
def supportuser(mailcow, db, tmpdir, mailcow_domain):
    addr = "pytest-supportuser-%s@%s" % (randint(0, 99999), mailcow_domain)
    tmpdir = Path(str(tmpdir))
    db_path = mailadm.bot.get_admbot_db_path(tmpdir.joinpath("supportuser.db"))
    ac = prepare_account(addr, mailcow, db_path)
    yield ac
    ac.shutdown()
    ac.wait_shutdown()
    mailcow.del_user_mailcow(addr)


@pytest.fixture
def mailcow_endpoint():
    if not os.environ.get("MAILCOW_ENDPOINT"):
        if os.environ.get("MAILCOW_TOKEN"):
            return "https://dc.develcow.de/api/v1/"
        pytest.skip("Please set the mailcow API URL with the environment variable MAILCOW_ENDPOINT")
    return os.environ.get("MAILCOW_ENDPOINT")


@pytest.fixture
def mailcow_auth():
    if not os.environ.get("MAILCOW_TOKEN"):
        pytest.skip("Please set a mailcow API Key with the environment variable MAILCOW_TOKEN")
    return {"X-API-Key": os.environ.get("MAILCOW_TOKEN")}


@pytest.fixture
def mailcow_domain():
    return os.environ.get("MAIL_DOMAIN", "x.testrun.org")


@pytest.fixture
def mailcow(db):
    with db.read_connection() as conn:
        return conn.get_mailcow_connection()


@pytest.fixture
def make_db(monkeypatch, mailcow_auth, mailcow_endpoint, mailcow_domain):
    def make_db(basedir, init=True):
        basedir = Path(str(basedir))
        db_path = basedir.joinpath("mailadm.db")
        db = mailadm.db.DB(db_path, debug=True)
        if init:
            db.init_config(
                mail_domain=mailcow_domain,
                web_endpoint="https://example.org/new_email",
                mailcow_endpoint=mailcow_endpoint,
                mailcow_token=mailcow_auth.get("X-API-Key"),
            )

        # re-route all queries for sysfiles to the tmpdir
        ttype = collections.namedtuple("pwentry", ["pw_name", "pw_dir", "pw_uid", "pw_gid"])

        def getpwnam(name):
            if name == "vmail":
                p = basedir.joinpath("path_vmaildir")
                if not p.exists():
                    p.mkdir()
            elif name == "mailadm":
                p = basedir
            else:
                raise KeyError("don't know user {!r}".format(name))
            return ttype(name, p, 10000, 10000)  # uid/gid should play no role for testing

        monkeypatch.setattr(pwd, "getpwnam", getpwnam)

        gtype = collections.namedtuple("grpentry", ["gr_name", "gr_mem"])

        def getgrnam(name):
            if name == "vmail":
                gr_mem = ["mailadm"]
            else:
                gr_mem = []
            return gtype(name, gr_mem)

        monkeypatch.setattr(grp, "getgrnam", getgrnam)

        return db

    return make_db



import random
import time

import mailadm
from mailadm.web import create_app_from_db_path


def test_new_user_random(request, db, monkeypatch, mailcow, mailcow_domain):
    token = "12319831923123"
    with db.write_transaction() as conn:
        conn.add_token(name="pytest", token=token, prefix="pytest.", expiry="1w")

    app = create_app_from_db_path(db.path)
    app.debug = True
    app = app.test_client()

    r = app.post("/?username=hello")
    assert r.status_code == 403
    assert r.json.get("reason") == "?t (token) parameter not specified"
    r = app.post("/?t=00000")
    assert r.status_code == 403
    assert r.json.get("reason") == "token 00000 is invalid"
    r = app.post("/?t=123123&username=hello")
    assert r.status_code == 403
    assert r.json.get("reason") == "token 123123 is invalid"

    # delete a@x.testrun.org and b@x.testrun.org in case earlier tests failed to clean them up
    user_a = "pytest.a@" + mailcow_domain
    user_b = "pytest.b@" + mailcow_domain

    def clean_up_test_users():
        mailcow.del_user_mailcow(user_a)
        mailcow.del_user_mailcow(user_b)

    request.addfinalizer(clean_up_test_users)

    chars = list("ab")

    def get_human_readable_id(*args, **kwargs):
        return random.choice(chars)

    monkeypatch.setattr(mailadm.util, "get_human_readable_id", get_human_readable_id)

    r = app.post("/?t=" + token)
    assert r.status_code == 200
    assert r.json["email"].endswith(mailcow_domain)
    assert r.json["password"]
    email = r.json["email"]
    assert email in [user_a, user_b]

    r2 = app.post("/?t=" + token)
    assert r2.status_code == 200
    assert r2.json["email"] != email
    assert r2.json["email"] in [user_a, user_b]

    r3 = app.post("/?t=" + token)
    assert r3.status_code == 409
    assert r3.json.get("reason") == "user already exists in mailcow"

    mailcow.del_user_mailcow(email)
    mailcow.del_user_mailcow(r2.json["email"])


def test_env(db, monkeypatch):
    monkeypatch.setenv("MAILADM_DB", str(db.path))
    from mailadm.app import app

    assert app.db.path == db.path


def test_user_in_db(db, mailcow):
    with db.write_transaction() as conn:
        token = conn.add_token("pytest:web", expiry="1w", token="1w_7wDioPeeXyZx96v", prefix="")
    app = create_app_from_db_path(db.path)
    app.debug = True
    app = app.test_client()

    r = app.post("/?t=" + token.token)
    assert r.status_code == 200
    assert r.json["password"]
    addr = r.json["email"]

    assert mailcow.get_user(addr)
    with db.read_connection() as conn:
        assert conn.get_user_by_addr(addr)

    with db.write_transaction() as conn:
        conn.delete_email_account(addr)


# we used to allow setting the username/password through the web
# but the code has been removed, let's keep the test around
def xxxtest_new_user_usermod(db, mailcow_domain):
    app = create_app_from_db_path(db.path)
    app.debug = True
    app = app.test_client()

    r = app.post("/?t=00000")
    assert r.status_code == 403

    r = app.post("/?t=123123123123123&username=hello")
    assert r.status_code == 200

    assert r.json["email"] == "hello@" + mailcow_domain
    assert len(r.json["password"]) >= 12

    now = time.time()
    r = app.post("/?t=123123123123123&username=hello2&password=l123123123123")
    assert r.status_code == 200
    assert r.json["email"] == "hello2@" + mailcow_domain
    assert r.json["password"] == "l123123123123"
    assert int(r.json["expires"]) > (now + 4 * 24 * 60 * 60)



from urllib.parse import quote

import requests as r

HTTP_TIMEOUT = 5


class MailcowConnection:
    """Class to manage requests to the mailcow instance.

    :param mailcow_endpoint: the URL to the mailcow API
    :param mailcow_token: the access token to the mailcow API
    """

    def __init__(self, mailcow_endpoint, mailcow_token):
        self.mailcow_endpoint = mailcow_endpoint
        self.auth = {"X-API-Key": mailcow_token}

    def add_user_mailcow(self, addr, password, token, quota=0):
        """HTTP Request to add a user to the mailcow instance.

        :param addr: the email address of the new account
        :param password: the password  of the new account
        :param token: the mailadm token used for account creation
        :param quota: the maximum mailbox storage in MB. default: unlimited
        """
        url = self.mailcow_endpoint + "add/mailbox"
        payload = {
            "local_part": addr.split("@")[0],
            "domain": addr.split("@")[1],
            "quota": quota,
            "password": password,
            "password2": password,
            "active": True,
            "force_pw_update": False,
            "tls_enforce_in": False,
            "tls_enforce_out": False,
            "tags": ["mailadm:" + token],
        }
        result = r.post(url, json=payload, headers=self.auth, timeout=HTTP_TIMEOUT)
        if type(result.json()) != list or result.json()[0].get("type") != "success":
            raise MailcowError(result.json())

    def del_user_mailcow(self, addr):
        """HTTP Request to delete a user from the mailcow instance.

        :param addr: the email account to be deleted
        """
        url = self.mailcow_endpoint + "delete/mailbox"
        result = r.post(url, json=[addr], headers=self.auth, timeout=HTTP_TIMEOUT)
        json = result.json()
        if not isinstance(json, list) or json[0].get("type" != "success"):
            raise MailcowError(json)

    def get_user(self, addr):
        """HTTP Request to get a specific mailcow user (not only mailadm-generated ones)."""
        url = self.mailcow_endpoint + "get/mailbox/" + quote(addr, safe="")
        result = r.get(url, headers=self.auth, timeout=HTTP_TIMEOUT)
        json = result.json()
        if json == {}:
            return None
        if type(json) == dict:
            if json.get("type") == "error":
                raise MailcowError(json)
            return MailcowUser(json)
        # some mailcow versions return all users, even if you only ask for a specific one:
        if type(json) == list:
            for user in [MailcowUser(user) for user in json]:
                if user.addr == addr:
                    return user

    def get_user_list(self):
        """HTTP Request to get all mailcow users (not only mailadm-generated ones)."""
        url = self.mailcow_endpoint + "get/mailbox/all"

        # Using larger timeout here than for other requests,
        # because some mailcow instances may have a large number of users.
        result = r.get(url, headers=self.auth, timeout=30)
        json = result.json()
        if json == {}:
            return []
        if type(json) == dict:
            if json.get("type") == "error":
                raise MailcowError(json)
        return [MailcowUser(user) for user in json]


class MailcowUser(object):
    def __init__(self, json):
        self.addr = json.get("username")
        self.quota = json.get("quota")
        self.last_login = json.get("last_imap_login")
        for tag in json.get("tags", []):
            if "mailadm:" in tag:
                self.token = tag.removeprefix("mailadm:")
                break


class MailcowError(Exception):
    """This is thrown if a Mailcow operation fails."""



import time
from random import randint

import deltachat
import mailadm
import pytest
import requests
from mailadm.conn import DBError, InvalidInputError
from mailadm.mailcow import MailcowError
from mailadm.util import parse_expiry_code


@pytest.fixture
def conn(db):
    with db.write_transaction() as conn:
        yield conn


def test_token_twice(conn):
    conn.add_token("pytest:burner1", expiry="1w", token="1w_7wDioPeeXyZx96v3", prefix="pp")
    with pytest.raises(DBError):
        conn.add_token("pytest:burner2", expiry="1w", token="1w_7wDioPeeXyZx96v3", prefix="xp")


def test_add_del_user(conn, mailcow):
    token = conn.add_token("pytest:burner1", expiry="1w", token="1w_7wDioPeeXyZx96v3", prefix="pp")
    usecount = conn.get_tokeninfo_by_name(token.name).usecount
    addr = conn.add_email_account_tries(token, tries=10).addr
    assert usecount + 1 == conn.get_tokeninfo_by_name(token.name).usecount

    assert mailcow.get_user(addr)
    assert conn.get_user_by_addr(addr)

    conn.delete_email_account(addr)

    with pytest.raises(TypeError):
        conn.get_user_by_addr(addr)
    assert not mailcow.get_user(addr)


def test_token_info(conn):
    conn.add_token("pytest:burner1", expiry="1w", token="1w_7wDioPeeXyZx96v3", prefix="pp")
    conn.add_token("pytest:burner2", expiry="10w", token="10w_7wDioPeeXyZx96v3", prefix="xp")

    assert conn.get_tokeninfo_by_token("1w_7wDio111111") is None
    ti = conn.get_tokeninfo_by_token("1w_7wDioPeeXyZx96v3")
    assert ti.expiry == "1w"
    assert ti.prefix == "pp"
    assert ti.name == "pytest:burner1"
    conn.del_token("pytest:burner2")
    assert not conn.get_tokeninfo_by_token("10w_7wDioPeeXyZx96v3")
    assert not conn.get_tokeninfo_by_name("pytest:burner2")


@pytest.mark.parametrize(
    ("localpart", "result"),
    [
        ("weirdinput.{}@test1@", False),
        ("weirdinput.{}@a", False),
        ("weirdinput.{}@x.testrun.org@test4@", False),
        ("weirdinput.{}/../../@", False),
        ("weirdinput.{}?test6@", False),
        ("weirdinput.{}#test7@", False),
        ("weirdinput./{}@", False),
        ("weirdinput.%2f{}@", False),
        ("weirdinput.{}\\test9\\\\@", False),
        ("weirdinput-{}@", True),
        ("weirdinput.{}", False),
        ("weirdinput+{}@", False),
        ("¹weirdinput.{}@", False),
        ("-weirdinput.{}@", False),
        ("_weirdinput.{}@", False),
        (".weirdinput.{}@", False),
        ("½weirdinput.{}@", False),
        ("weirdinput_{}@", True),
    ],
)
def test_is_email_valid(conn, localpart, result, mailcow_domain):
    addr = localpart.format(randint(0, 99999)) + mailcow_domain
    assert conn.is_valid_email(addr) == result


@pytest.mark.parametrize(
    ("token_name", "result"),
    [
        ("../test", False),
        ("../../test", False),
        ("../../abc/../test", False),
        (".abc/../test/fixed", False),
        ("/test/foo", False),
        ("./test/baz", False),
        (".testbaz", False),
        ("test%baz", False),
        ("test?secret=asdf", False),
        ("test#somewhere", False),
        ("baz123-._@", True),
    ],
)
def test_token_sanitization(conn, token_name, result):
    if result:
        conn.add_token(token_name, expiry="1w", token="1w_7wDioPeeXyZx96va3", prefix="pp")
    else:
        with pytest.raises(InvalidInputError):
            conn.add_token(token_name, expiry="1w", token="1w_7wDioPeeXyZx96va3", prefix="pp")


def test_email_tmp_gen(conn, mailcow):
    conn.add_token("pytest:burner1", expiry="1w", token="1w_7wDioPeeXyZx96v3", prefix="tmp.")
    token_info = conn.get_tokeninfo_by_name("pytest:burner1")
    user_info = conn.add_email_account(token_info=token_info)

    assert user_info.token_name == "pytest:burner1"
    localpart, domain = user_info.addr.split("@")
    assert localpart.startswith("tmp.")
    assert domain == conn.config.mail_domain

    username = localpart[4:]
    assert len(username) == 5
    for c in username:
        assert c in "2345789acdefghjkmnpqrstuvwxyz"

    mailcow.del_user_mailcow(user_info.addr)


def test_adduser_mailcow_error(db):
    """Test that DB doesn't change if mailcow doesn't work"""
    with db.write_transaction() as conn:
        token_info = conn.add_token(
            "pytest:burner1",
            expiry="1w",
            token="1w_7wDioPeeXyZx96v3",
            prefix="tmp.",
            maxuse=1,
        )

    with db.write_transaction() as conn:
        conn.set_config("mailcow_token", "wrong")
        with pytest.raises(MailcowError):
            conn.add_email_account(token_info)

    with db.write_transaction() as conn:
        token_info = conn.get_tokeninfo_by_name(token_info.name)
        token_info.check_exhausted()
        assert conn.get_user_list(token=token_info.name) == []


def test_adduser_db_error(conn, monkeypatch, mailcow_domain):
    """Test that no mailcow user is created if there is a DB error"""
    token_info = conn.add_token(
        "pytest:burner1",
        expiry="1w",
        token="1w_7wDioPeeXyZx96v3",
        prefix="tmp.",
    )
    addr = "pytest.%s@%s" % (randint(0, 99999), mailcow_domain)

    def add_user_db(*args, **kwargs):
        raise DBError

    monkeypatch.setattr(mailadm.conn.Connection, "add_user_db", add_user_db)

    with pytest.raises(DBError):
        conn.add_email_account(token_info, addr=addr)

    url = "%sget/mailbox/%s" % (conn.config.mailcow_endpoint, addr)
    auth = {"X-API-Key": conn.config.mailcow_token}
    result = requests.get(url, headers=auth)
    assert result.status_code == 200
    if result.json() is not {} and type(result.json()) == list:
        for user in result.json():
            assert user["username"] != addr


def test_adduser_mailcow_exists(conn, mailcow, mailcow_domain):
    """Test that no user is created if Mailcow user already exists"""
    token_info = conn.add_token("pytest:burner1", expiry="1w", token="1w_7wDioPeeXyZx", prefix="p.")
    addr = "%s@%s" % (randint(0, 99999), mailcow_domain)

    mailcow.add_user_mailcow(addr, "asdf1234", token_info.name)
    with pytest.raises(MailcowError):
        conn.add_email_account(token_info, addr=addr)
    for user in conn.get_user_list():
        assert user.token_name == "created in mailcow"

    mailcow.del_user_mailcow(addr)


@pytest.mark.parametrize(
    ("expiry", "time_overdue", "time_since_last_login", "to_be_pruned"),
    [
        ("1h", "1h", "0s", True),
        ("1s", "1s", "0s", True),
        ("5w", "2w", "1w", False),
        ("2y", "2y", "20d", False),
        ("2y", "1d", "30w", True),
        ("90d", "30d", "2d", False),
        ("30d", "2y", "1w", False),
    ],
)
def test_soft_expiry(
    conn,
    mailcow_domain,
    tmp_db_path,
    monkeypatch,
    expiry,
    time_overdue,
    time_since_last_login,
    to_be_pruned,
    mailcow,
):
    """Test whether a user expires depending on last login"""
    to_expire = time.time() - parse_expiry_code(expiry) - parse_expiry_code(time_overdue)
    # create an account which is living for longer than its expiry time so far
    with monkeypatch.context() as m:
        m.setattr(time, "time", lambda: to_expire)
        token_info = conn.add_token("pytest:soft", expiry=expiry, token="1w_7wDioPeex", prefix="p.")
        addr = "pytest.%s@%s" % (randint(0, 99999), mailcow_domain)
        mailadm_user = conn.add_email_account(token_info, addr=addr)

    # login with IMAP
    ac = deltachat.Account(tmp_db_path)
    ac.run_account(addr=mailadm_user.addr, password=mailadm_user.password)

    # make mailadm think that it's in the future and user hasn't been logged in for a while
    future = time.time() + parse_expiry_code(time_since_last_login)

    expired_users = conn.get_expired_users(future)
    expired = mailadm_user.addr in [user.addr for user in expired_users]
    assert expired == to_be_pruned

    mailcow.del_user_mailcow(addr)


def test_delete_user_mailcow_missing(conn, mailcow, mailcow_domain):
    """Test if a mailadm user is deleted successfully if mailcow user is already missing"""
    token_info = conn.add_token("pytest:burner1", expiry="1w", token="1w_7wDioPeeXyZx", prefix="p.")
    addr = "pytest.%s@%s" % (randint(0, 99999), mailcow_domain)

    conn.add_email_account(token_info, addr=addr)
    mailcow.del_user_mailcow(addr)
    conn.delete_email_account(addr)


def test_db_version(conn):
    version = conn.get_dbversion()
    assert type(version) == int



import logging
import mimetypes
import os
import sqlite3
import sys
import time
from threading import Event

import deltachat
from deltachat import account_hookimpl
from deltachat.capi import lib as dclib

from mailadm.commands import add_token, add_user, list_tokens, prune, qr_from_token
from mailadm.db import DB, get_db_path


class SetupPlugin:
    def __init__(self, admingrpid):
        self.member_added = Event()
        self.admingrpid = admingrpid
        self.message_sent = Event()

    @account_hookimpl
    def ac_member_added(self, chat: deltachat.Chat, contact, actor, message):
        if chat.id == self.admingrpid and chat.num_contacts() == 2:
            self.member_added.set()

    @account_hookimpl
    def ac_message_delivered(self, message: deltachat.Message):
        if not message.is_system_message():
            self.message_sent.set()


class AdmBot:
    def __init__(self, db: DB, account: deltachat.Account):
        self.db = db
        self.account = account
        with self.db.read_connection() as conn:
            config = conn.config
            self.admingrpid = int(config.admingrpid)
            self.admingroup = account.get_chat_by_id(self.admingrpid)
            self.mail_domain = config.mail_domain

    @account_hookimpl
    def ac_incoming_message(self, message: deltachat.Message):
        """This method is called on every incoming message and decides what to do with it."""
        logging.info("new message from %s: %s", message.get_sender_contact().addr, message.text)
        if self.is_admin_group_message(message):
            if message.text.startswith("/"):
                logging.info("%s seems to be a command.", message.text)
                self.handle_command(message)
            else:
                logging.debug("ignoring message, it's just admins discussing in the admin group")
        elif self.is_support_group(message.chat):
            if message.quote:
                if message.quote.get_sender_contact().addr == self.account.get_config("addr"):
                    self.forward_reply_to_support_user(message)
            elif message.text.startswith("/"):
                logging.info("ignoring command, it wasn't given in the admin group")
                message.chat.send_text("Sorry, I only take commands in the admin group.")
            else:
                logging.debug("ignoring message, it's just admins discussing in a support group")
        else:
            chat = message.create_chat()
            if chat.is_group():
                logging.info(
                    "%s added me to a group, I'm leaving it.",
                    message.get_sender_contact().addr,
                )
                chat.send_text("Sorry, you can not contact me in groups. Please use a 1:1 chat.")
                chat.remove_contact(self.account.get_self_contact())  # leave group
            elif message.text[0:5] == "/help":
                chat.send_text("You can use this chat to talk to the admins.")
            elif message.text[0] == "/":
                logging.info("ignoring command, it wasn't given in the admin group")
                chat.send_text("Sorry, I only take commands in the admin group.")
            else:
                logging.info("forwarding the message to a support group.")
                self.forward_to_support_group(message)

    def is_support_group(self, chat: deltachat.Chat) -> bool:
        """Checks whether the group was created by the bot."""
        return (
            chat.is_group()
            and chat.get_messages()[0].get_sender_contact() == self.account.get_self_contact()
        )

    def forward_to_support_group(self, message: deltachat.Message):
        """forward a support request to a support group; create one if it doesn't exist yet."""
        support_user = message.get_sender_contact().addr
        admins = self.admingroup.get_contacts()
        admins.remove(self.account.get_self_contact())
        group_name = support_user + " support group"
        for chat in self.account.get_chats():
            if chat.get_name() == group_name:
                support_group = chat
                break
        else:
            logging.info("creating new support group: '%s'", group_name)
            support_group = self.account.create_group_chat(group_name, admins)
            support_group.set_profile_image("assets/avatar.jpg")
        message.set_override_sender_name(support_user)
        support_group.send_msg(message)

    def forward_reply_to_support_user(self, message: deltachat.Message):
        """an admin replied in a support group; forward their reply to the user."""
        recipient = message.quote.override_sender_name
        logging.info("I'm forwarding the admin reply to the support user %s.", recipient)
        chat = self.account.create_chat(recipient)
        chat.send_msg(message)

    def is_admin_group_message(self, command: deltachat.Message):
        """Checks whether the incoming message was in the admin group."""
        if command.chat.is_group() and self.admingrpid == command.chat.id:
            if (
                command.chat.is_protected()
                and command.is_encrypted()
                and int(command.chat.num_contacts()) >= 2
            ):
                if command.get_sender_contact() in command.chat.get_contacts():
                    return True
                else:
                    logging.info(
                        "%s is not allowed to give commands to mailadm.",
                        command.get_sender_contact(),
                    )
            else:
                logging.info(
                    "The admin group is broken. Try `mailadm setup-bot`. Group ID: %s",
                    str(self.admingrpid),
                )
                raise ValueError
        else:
            return False

    def handle_command(self, message: deltachat.Message):
        """execute the command and reply to the admin."""
        arguments = message.text.split(" ")
        image_path = None

        if arguments[0] == "/add-token":
            text, image_path = self.add_token(arguments)

        elif arguments[0] == "/gen-qr":
            text, image_path = self.gen_qr(arguments)

        elif arguments[0] == "/add-user":
            text = self.add_user(arguments)

        elif arguments[0] == "/list-users":
            text = self.list_users(arguments)

        elif arguments[0] == "/list-tokens":
            text = list_tokens(self.db)

        else:
            text = (
                "/add-user addr password token\n"
                "/add-token name expiry maxuse (prefix)\n"
                "/gen-qr token\n"
                "/list-users (token)\n"
                "/list-tokens"
            )

        if image_path:
            msg = deltachat.Message.new_empty(self.account, "image")
            mime_type = mimetypes.guess_type(image_path)[0]
            msg.set_file(image_path, mime_type)
        else:
            msg = deltachat.Message.new_empty(self.account, "text")
        msg.set_text(text)
        msg.quote = message
        sent_id = dclib.dc_send_msg(self.account._dc_context, self.admingroup.id, msg._dc_msg)
        assert sent_id == msg.id

    def add_token(self, arguments: [str]):
        """add a token via bot command"""
        if len(arguments) == 4:
            arguments.append("")  # add empty prefix
        if len(arguments) < 4:
            result = {
                "status": "error",
                "message": "Sorry, you need to tell me more precisely what you want. For "
                "example:\n\n/add-token oneweek 1w 50\n\nThis would create a token"
                " which creates up to 50 accounts which each are valid for one "
                "week.",
            }
        else:
            result = add_token(
                self.db,
                name=arguments[1],
                expiry=arguments[2],
                maxuse=arguments[3],
                prefix=arguments[4],
                token=None,
            )
        if result["status"] == "error":
            return "ERROR: " + result.get("message"), None
        text = result.get("message")
        fn = qr_from_token(self.db, arguments[1])["filename"]
        return text, fn

    def gen_qr(self, arguments: [str]):
        """generate a QR code via bot command"""
        if len(arguments) != 2:
            return "Sorry, which token do you want a QR code for?", None
        else:
            return "", qr_from_token(self.db, tokenname=arguments[1]).get("filename")

    def add_user(self, arguments: [str]):
        """add a user via bot command"""
        if len(arguments) < 4:
            try:
                with self.db.read_connection() as conn:
                    token_name = conn.get_token_list()[0]
            except IndexError:
                return "You need to create a token with /add-token first."
            result = {
                "status": "error",
                "message": "Sorry, you need to tell me more precisely what you want. For "
                "example:\n\n/add-user test@%s p4$$w0rd %s\n\nThis would "
                "create a user with the '%s' token and the password "
                "'p4$$w0rd'." % (self.mail_domain, token_name, token_name),
            }
        else:
            result = add_user(self.db, addr=arguments[1], password=arguments[2], token=arguments[3])
        if result.get("status") == "success":
            user = result.get("message")
            return "successfully created %s with password %s" % (user.addr, user.password)
        else:
            return result.get("message")

    def list_users(self, arguments: [str]):
        """list users per bot command"""
        token = arguments[1] if len(arguments) > 1 else None
        with self.db.read_connection() as conn:
            users = conn.get_user_list(token=token)
        lines = ["%s [%s]" % (user.addr, user.token_name) for user in users]
        return "\n".join(lines)


def get_admbot_db_path(db_path=None):
    if not db_path:
        db_path = os.environ.get("ADMBOT_DB", "/mailadm/docker-data/admbot.db")
    try:
        sqlite3.connect(db_path)
    except sqlite3.OperationalError:
        raise RuntimeError("admbot.db not found: ADMBOT_DB not set")
    return db_path


def main(mailadm_db, admbot_db_path):
    try:
        ac = deltachat.Account(admbot_db_path)
        with mailadm_db.read_connection() as conn:
            if "admingrpid" not in [item[0] for item in conn.get_config_items()]:
                # the file=sys.stderr seems to be necessary so the output is shown in `docker logs`
                print(
                    "To complete the mailadm setup, please run: mailadm setup-bot",
                    file=sys.stderr,
                )
                os._exit(1)
            displayname = conn.config.mail_domain + " administration"
        ac.set_avatar("assets/avatar.jpg")
        ac.run_account(account_plugins=[AdmBot(mailadm_db, ac)], show_ffi=True)
        ac.set_config("mvbox_move", "1")
        ac.set_config("show_emails", "2")
        ac.set_config("displayname", displayname)
        while 1:
            for logmsg in prune(mailadm_db).get("message"):
                logging.info("%s", logmsg)
            for _second in range(0, 600):
                if not ac._event_thread.is_alive():
                    logging.error("dc core event thread died, exiting now")
                    os._exit(1)
                time.sleep(1)
    except Exception:
        logging.exception("bot received an unexpected error, exiting now")
        os._exit(1)


if __name__ == "__main__":
    mailadm_db = DB(get_db_path())
    admbot_db_path = get_admbot_db_path()
    main(mailadm_db, admbot_db_path)



"""
help gunicorn and other WSGI servers to instantiate a web instance of mailadm
"""
import threading

from mailadm.bot import get_admbot_db_path
from mailadm.bot import main as run_bot

from .db import DB, get_db_path
from .web import create_app_from_db_path


def init_threads():
    botthread = threading.Thread(
        target=run_bot,
        args=(DB(get_db_path()), get_admbot_db_path()),
        daemon=True,
        name="bot",
    )
    botthread.start()


app = create_app_from_db_path()



import logging
import sqlite3
import time

import mailadm.util

from .mailcow import MailcowConnection, MailcowError


class DBError(Exception):
    """error during an operation on the database."""


class TokenExhaustedError(DBError):
    """A token has reached its max-use limit."""


class UserNotFoundError(DBError):
    """user not found in database."""


class InvalidInputError(DBError):
    """Raised when user-specified input was invalid"""


class Connection:
    def __init__(self, sqlconn, path, write):
        self._sqlconn = sqlconn
        self.path_mailadm_db = path
        self._write = write

    def log(self, msg):
        logging.info("%s", msg)

    def close(self):
        self._sqlconn.close()

    def commit(self):
        self._sqlconn.commit()

    def rollback(self):
        self._sqlconn.rollback()

    def execute(self, query, params=()):
        cur = self.cursor()
        try:
            cur.execute(query, params)
        except sqlite3.IntegrityError as e:
            raise DBError(e)
        return cur

    def cursor(self):
        return self._sqlconn.cursor()

    #
    # configuration and meta information
    #
    def get_dbversion(self):
        q = "SELECT value from config WHERE name='dbversion'"
        c = self._sqlconn.cursor()
        try:
            return int(c.execute(q).fetchone()[0])
        except sqlite3.OperationalError:
            return None

    @property
    def config(self):
        items = self.get_config_items()
        if items:
            d = dict(items)
            # remove deprecated config keys
            try:
                del d["vmail_user"]
            except KeyError:
                pass
            try:
                del d["path_virtual_mailboxes"]
            except KeyError:
                pass
            return Config(**d)

    def is_initialized(self):
        items = self.get_config_items()
        return len(items) > 1

    def get_config_items(self):
        q = "SELECT name, value from config"
        c = self._sqlconn.cursor()
        try:
            return c.execute(q).fetchall()
        except sqlite3.OperationalError:
            return None

    def set_config(self, name, value):
        ok = [
            "dbversion",
            "mail_domain",
            "web_endpoint",
            "mailcow_endpoint",
            "mailcow_token",
            "admingrpid",
        ]
        assert name in ok, name
        q = "INSERT OR REPLACE INTO config (name, value) VALUES (?, ?)"
        self.cursor().execute(q, (name, value)).fetchone()
        return value

    #
    # token management
    #

    def get_token_list(self):
        q = "SELECT name from tokens"
        return [x[0] for x in self.execute(q).fetchall()]

    def add_token(self, name, token, expiry, prefix, maxuse=50):
        if "/" in name or "#" in name or "?" in name or "%" in name:
            raise InvalidInputError("no /, ?, %, or # allowed in the token name")
        if name[0] == ".":
            raise InvalidInputError("token name can't start with a dot (.)")
        q = "INSERT INTO tokens (name, token, prefix, expiry, maxuse) VALUES (?, ?, ?, ?, ?)"
        self.execute(q, (name, token, prefix, expiry, int(maxuse)))
        self.log("added token {!r}".format(name))
        return self.get_tokeninfo_by_name(name)

    def mod_token(self, name, expiry=None, prefix=None, maxuse=None):
        token_info = self.get_tokeninfo_by_name(name)
        expiry = expiry if expiry is not None else token_info.expiry
        maxuse = maxuse if maxuse is not None else token_info.maxuse
        prefix = prefix if prefix is not None else token_info.prefix
        q = "REPLACE INTO tokens (name, token, prefix, expiry, maxuse) VALUES (?, ?, ?, ?, ?)"
        self.execute(q, (name, token_info.token, prefix, expiry, maxuse))
        self.log("modified token {!r}".format(name))
        return self.get_tokeninfo_by_name(name)

    def del_token(self, name):
        q = "DELETE FROM tokens WHERE name=?"
        c = self.cursor()
        c.execute(q, (name,))
        if c.rowcount == 0:
            raise ValueError("token {!r} does not exist".format(name))
        self.log("deleted token {!r}".format(name))

    def get_tokeninfo_by_name(self, name):
        q = TokenInfo._select_token_columns + "WHERE name = ?"
        res = self.execute(q, (name,)).fetchone()
        if res is not None:
            return TokenInfo(self.config, *res)

    def get_tokeninfo_by_token(self, token):
        q = TokenInfo._select_token_columns + "WHERE token=?"
        res = self.execute(q, (token,)).fetchone()
        if res is not None:
            return TokenInfo(self.config, *res)

    def get_tokeninfo_by_addr(self, addr):
        if not addr.endswith(self.config.mail_domain):
            raise ValueError(
                "addr {!r} does not use mail domain {!r}".format(addr, self.config.mail_domain),
            )
        q = TokenInfo._select_token_columns
        for res in self.execute(q).fetchall():
            token_info = TokenInfo(self.config, *res)
            if addr.startswith(token_info.prefix):
                return token_info

    #
    # user management
    #

    def add_email_account_tries(self, token_info, addr=None, password=None, tries=1):
        """Try to add an email account."""
        for i in range(1, tries + 1):
            logging.info("Try %d to create an account", i)
            try:
                return self.add_email_account(token_info, addr=addr, password=password)
            except (MailcowError, DBError):
                if i >= tries:
                    raise

    def add_email_account(self, token_info, addr=None, password=None):
        """Add an email account to the mailcow server & mailadm

        :param token_info: the token which authorizes the new user creation
        :param addr: email address for the new account; randomly generated if omitted
        :param password: password for the new account; randomly generated if omitted
        :return: a UserInfo object with the database information about the new user, plus password
        """
        token_info.check_exhausted()
        if password is None:
            password = mailadm.util.gen_password()
        if addr is None:
            rand_part = mailadm.util.get_human_readable_id()
            username = "{}{}".format(token_info.prefix, rand_part)
            addr = "{}@{}".format(username, self.config.mail_domain)

        if not self.is_valid_email(addr):
            raise InvalidInputError("not a valid email address")

        # first check that mailcow doesn't have a user with that name already:
        if self.get_mailcow_connection().get_user(addr):
            raise MailcowError("account does already exist")

        self.add_user_db(
            addr=addr,
            date=int(time.time()),
            ttl=token_info.get_expiry_seconds(),
            token_name=token_info.name,
        )

        self.log("added addr {!r} with token {!r}".format(addr, token_info.name))

        user_info = self.get_user_by_addr(addr)
        user_info.password = password

        # seems that everything is fine so far, so let's invoke mailcow:
        self.get_mailcow_connection().add_user_mailcow(addr, password, token_info.name)

        return user_info

    def delete_email_account(self, addr):
        """Delete an email account from the mailcow server & mailadm.

        :param addr: the email address of the account which is to be deleted.
        """
        self.get_mailcow_connection().del_user_mailcow(addr)
        self.del_user_db(addr)

    def add_user_db(self, addr, date, ttl, token_name):
        self.execute("PRAGMA foreign_keys=on;")

        q = """INSERT INTO users (addr, date, ttl, token_name)
               VALUES (?, ?, ?, ?)"""
        self.execute(q, (addr, date, ttl, token_name))
        self.execute("UPDATE tokens SET usecount = usecount + 1 WHERE name=?", (token_name,))

    def del_user_db(self, addr):
        q = "DELETE FROM users WHERE addr=?"
        c = self.execute(q, (addr,))
        if c.rowcount == 0:
            raise UserNotFoundError("addr {!r} does not exist".format(addr))
        self.log("deleted user {!r}".format(addr))

    def is_valid_email(self, addr):
        if not addr.endswith("@" + self.config.mail_domain):
            logging.error("address %s doesn't end with @x.testrun.org", addr)
            return False
        if not addr.count("@") == 1:
            logging.error("address %s doesn't have exactly one @", addr)
            return False
        if not addr[0].isalnum():
            logging.error("address %s doesn't start alphanumeric", addr)
            return False
        chars_allowed = "abcdefghijklmnopqrstuvwxyz-ABCDEFGHIJKLMNOPQRSTUVWXYZ.0123456789_@"
        for char in addr:
            if char not in chars_allowed:
                logging.error("letter %s in address %s is not in %s", char, addr, chars_allowed)
                return False
        return True

    def get_user_by_addr(self, addr):
        q = UserInfo._select_user_columns + "WHERE addr = ?"
        args = self._sqlconn.execute(q, (addr,)).fetchone()
        return UserInfo(*args)

    def get_expired_users(self, sysdate):
        q = UserInfo._select_user_columns + "WHERE (date + ttl) < ?"
        overdue_users = []
        expired_users = []
        for args in self._sqlconn.execute(q, (sysdate,)).fetchall():
            overdue_users.append(UserInfo(*args))
        for user in overdue_users:
            # expire users who were supposed to live less than 27 days
            if user.ttl < mailadm.util.parse_expiry_code("27d"):
                expired_users.append(user)
                continue
            last_login = self.get_mailcow_connection().get_user(user.addr).last_login
            # expire users who weren't online for longer than 25% of their TTL:
            if sysdate - last_login > user.ttl * 0.25:
                expired_users.append(user)
        return expired_users

    def get_user_list(self, token=None):
        q = UserInfo._select_user_columns
        args = []
        if token is not None:
            q += "WHERE token_name=?"
            args.append(token)
        dbusers = [UserInfo(*args) for args in self._sqlconn.execute(q, args).fetchall()]
        try:
            mcusers = self.get_mailcow_connection().get_user_list()
            if not token:
                for mcuser in mcusers:
                    if mcuser.addr not in [dbuser.addr for dbuser in dbusers]:
                        dbusers.append(UserInfo(mcuser.addr, 0, 0, "created in mailcow"))
            for dbuser in dbusers:
                if dbuser.addr not in [mcuser.addr for mcuser in mcusers]:
                    dbuser.token_name = "WARNING: does not exist in mailcow"
        except MailcowError as e:
            self.log("Can't check mailcow users: " + str(e))
        return dbusers

    def get_mailcow_connection(self) -> MailcowConnection:
        return MailcowConnection(self.config.mailcow_endpoint, self.config.mailcow_token)


class TokenInfo:
    _select_token_columns = "SELECT name, token, expiry, prefix, maxuse, usecount from tokens\n"

    def __init__(self, config, name, token, expiry, prefix, maxuse, usecount):
        self.config = config
        self.name = name
        self.token = token
        self.expiry = expiry
        self.prefix = prefix
        self.maxuse = maxuse
        self.usecount = usecount

    def get_maxdays(self):
        return mailadm.util.parse_expiry_code(self.expiry) / (24 * 60 * 60)

    def get_expiry_seconds(self):
        return mailadm.util.parse_expiry_code(self.expiry)

    def get_web_url(self):
        return "{web}?t={token}&n={name}".format(
            web=self.config.web_endpoint,
            token=self.token,
            name=self.name,
        )

    def get_qr_uri(self):
        return "DCACCOUNT:" + self.get_web_url()

    def check_exhausted(self):
        """Check if a token can still create email accounts."""
        if self.usecount >= self.maxuse:
            raise TokenExhaustedError


class UserInfo:
    _select_user_columns = "SELECT addr, date, ttl, token_name from users\n"

    def __init__(self, addr, date, ttl, token_name):
        self.addr = addr
        self.date = date
        self.ttl = ttl
        self.token_name = token_name


class Config:
    """The mailadm config.

    :param mail_domain: the domain of the mailserver
    :param web_endpoint: the web endpoint of mailadm's web interface
    :param dbversion: the version of the mailadm database schema
    :param mailcow_endpoint: the URL to the mailcow API
    :param mailcow_token: the token to authenticate with the mailcow API
    :param admingrpid: the ID of the admin group
    """

    def __init__(
        self,
        mail_domain,
        web_endpoint,
        dbversion,
        mailcow_endpoint,
        mailcow_token,
        admingrpid=None,
    ):
        self.mail_domain = mail_domain
        self.web_endpoint = web_endpoint
        self.dbversion = dbversion
        self.mailcow_endpoint = mailcow_endpoint
        self.mailcow_token = mailcow_token
        self.admingrpid = admingrpid



import os

import pkg_resources
import qrcode
from PIL import Image, ImageDraw, ImageFont


def gen_qr(config, token_info):
    info = "{prefix}******@{domain} {expiry}\n".format(
        domain=config.mail_domain,
        prefix=token_info.prefix,
        expiry=token_info.expiry,
    )

    steps = (
        "1. Install https://get.delta.chat\n"
        "2. From setup screen scan above QR code\n"
        "3. Choose nickname & avatar\n"
        "+ chat with any e-mail address ...\n"
    )

    # load QR code
    url = token_info.get_qr_uri()
    qr = qrcode.QRCode(
        version=1,
        error_correction=qrcode.constants.ERROR_CORRECT_H,
        box_size=1,
        border=1,
    )
    qr.add_data(url)
    qr.make(fit=True)
    qr_img = qr.make_image(fill_color="black", back_color="white")

    # paint all elements
    ttf_path = pkg_resources.resource_filename("mailadm", "data/opensans-regular.ttf")
    logo_red_path = pkg_resources.resource_filename("mailadm", "data/delta-chat-red.png")

    assert os.path.exists(ttf_path), ttf_path
    font_size = 16
    font = ImageFont.truetype(font=ttf_path, size=font_size)

    num_lines = (info + steps).count("\n") + 2

    size = width = 384
    qr_padding = 6
    text_margin_right = 12
    text_height = font_size * num_lines
    height = size + text_height + qr_padding * 2

    image = Image.new("RGBA", (width, height), "white")

    draw = ImageDraw.Draw(image)

    qr_final_size = width - (qr_padding * 2)

    # draw text
    info_pos = (width - font.getsize(info.strip())[0]) // 2
    draw.multiline_text(
        (info_pos, size - qr_padding // 2),
        info,
        font=font,
        fill="red",
        align="right",
    )
    draw.multiline_text(
        (text_margin_right, height - text_height + font_size * 1.0),
        steps,
        font=font,
        fill="black",
        align="left",
    )

    # paste QR code
    image.paste(
        qr_img.resize((qr_final_size, qr_final_size), resample=Image.NEAREST),
        (qr_padding, qr_padding),
    )

    # paste black and white logo
    # logo_bw_path = pkg_resources.resource_filename('mailadm', 'data/delta-chat-bw.png')
    # logo_img = Image.open(logo_bw_path)
    # logo = logo_img.resize((logo_width, logo_width), resample=Image.NEAREST)
    # image.paste(logo, (0, qr_final_size + qr_padding), mask=logo)

    # red background delta logo
    logo2_img = Image.open(logo_red_path)
    logo2_width = int(size / 6)
    logo2 = logo2_img.resize((logo2_width, logo2_width), resample=Image.NEAREST)
    pos = int((size / 2) - (logo2_width / 2))
    image.paste(logo2, (pos, pos), mask=logo2)

    return image



import base64
import random
import sys


def gen_password():
    with open("/dev/urandom", "rb") as f:
        s = f.read(21)
    return base64.b64encode(s).decode("ascii")[:12]


def get_human_readable_id(len=5, chars="2345789acdefghjkmnpqrstuvwxyz"):
    return "".join(random.choice(chars) for i in range(len))


def parse_expiry_code(code):
    if code == "never":
        return sys.maxsize

    if len(code) < 2:
        raise ValueError("expiry codes are at least 2 characters")
    val = int(code[:-1])
    c = code[-1]
    if c == "y":
        return val * 365 * 24 * 60 * 60
    elif c == "w":
        return val * 7 * 24 * 60 * 60
    elif c == "d":
        return val * 24 * 60 * 60
    elif c == "h":
        return val * 60 * 60
    elif c == "s":
        return val
    else:
        raise ValueError(c + " is not a valid time unit. Try [y]ears, [w]eeks, [d]ays, or [h]ours")



"""
script implementation of
https://github.com/codespeaknet/sysadmin/blob/master/docs/postfix-virtual-domains.rst#add-a-virtual-mailbox

"""

from __future__ import print_function

import sys

import click
import qrcode
from click import style
from deltachat import Account, account_hookimpl
from deltachat.events import FFIEventLogger
from deltachat.tracker import ConfigureFailed

import mailadm
import mailadm.commands
import mailadm.db
import mailadm.util

from .bot import SetupPlugin, get_admbot_db_path
from .conn import DBError, UserInfo
from .mailcow import MailcowError

option_dryrun = click.option(
    "-n",
    "--dryrun",
    is_flag=True,
    help="don't change any files, only show what would be changed.",
)


@click.command(cls=click.Group, context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option()
@click.pass_context
def mailadm_main(context):
    """e-mail account creation admin tool and web service."""


def get_mailadm_db(ctx, show=False, fail_missing_config=True):
    try:
        db_path = mailadm.db.get_db_path()
    except RuntimeError as e:
        ctx.fail(e.args)

    try:
        db = mailadm.db.DB(db_path)
    except DBError as e:
        ctx.fail(str(e))

    if show:
        click.secho("using db: {}".format(db_path), file=sys.stderr)
    if fail_missing_config:
        with db.read_connection() as conn:
            if not conn.is_initialized():
                ctx.fail("database not initialized, use 'init' subcommand to do so")
    return db


def create_bot_account(ctx, email: str, password=None) -> (str, str):
    """Creates a mailcow account to use for the bot.

    :param email: the email address to be used for the bot account
    :param password: you can set a custom password via CLI, auto-generated if ommitted
    :return email, password: the credentials for the bot account
    """
    mailadmdb = get_mailadm_db(ctx)
    with mailadmdb.read_connection() as rconn:
        mc = rconn.get_mailcow_connection()
        if not password:
            password = mailadm.util.gen_password()
        try:
            mc.add_user_mailcow(email, password, "bot")
        except MailcowError as e:
            if "object_exists" in str(e):
                ctx.fail(
                    "%s already exists; delete the account in mailcow or specify "
                    "credentials with --email and --password." % (email,),
                )
            else:
                raise
        print("New account %s created as bot account." % (email,))
    return email, password


@click.command()
@click.option("--email", type=str, default=None, help="name of email")
@click.option("--password", type=str, default=None, help="name of password")
@click.option("--show-ffi", is_flag=True, help="show low level ffi events")
@click.pass_context
@account_hookimpl
def setup_bot(ctx, email, password, show_ffi):
    """initialize the deltachat bot as an alternative command interface.

    :param ctx: the click object passing the CLI environment
    :param email: the email account the deltachat bot will use for receiving commands
    :param password: the password to the bot's email account
    :param db: the path to the deltachat database of the bot - NOT the path to the mailadm database!
    :param show_ffi: show low level ffi events
    """
    admbot_db = get_admbot_db_path()
    ac = Account(admbot_db)
    if show_ffi:
        ac.add_account_plugin(FFIEventLogger(ac))

    mailadmdb = get_mailadm_db(ctx)
    with mailadmdb.read_connection() as rconn:
        mail_domain = rconn.config.mail_domain
        admingrpid_old = rconn.config.admingrpid

    if not ac.is_configured():
        if email and not password:
            if email.split("@")[1] == mail_domain:
                print("--password not specified, creating account automatically... ")
                email, password = create_bot_account(ctx, email)
            else:
                ctx.fail(
                    "You need to provide --password if you want to use an existing account "
                    "for the mailadm bot.",
                )
        elif not email and not password:
            print("--email and --password not specified, creating account automatically... ")
            email = "mailadm@" + mail_domain
            email, password = create_bot_account(ctx, email, password=password)
        elif not email and password:
            ctx.fail("Please also provide --email to use an email account for the mailadm bot.")
    if email:
        ac.set_config("addr", email)
    if password:
        ac.set_config("mail_pw", password)
    ac.set_config("mvbox_move", "1")
    ac.set_config("sentbox_watch", "0")
    ac.set_config("bot", "1")
    configtracker = ac.configure()
    try:
        configtracker.wait_finish()
    except ConfigureFailed as e:
        ctx.fail("Authentication Failed: " + str(e))

    ac.start_io()

    chat = ac.create_group_chat("Admin group on {}".format(mail_domain), contacts=[], verified=True)

    setupplugin = SetupPlugin(chat.id)
    ac.add_account_plugin(setupplugin)

    chatinvite = chat.get_join_qr()
    qr = qrcode.QRCode()
    qr.add_data(chatinvite)
    print("\nPlease scan this qr code to join a verified admin group chat:\n\n")
    qr.print_ascii(invert=True)
    print("\nAlternatively, copy-paste this invite to your Delta Chat desktop client:", chatinvite)

    print("\nWaiting until you join the chat")
    sys.stdout.flush()  # flush stdout to actually show the messages above
    setupplugin.member_added.wait()
    setupplugin.message_sent.clear()

    chat.send_text(
        "Welcome to the Admin group on %s! Type /help to see the existing commands."
        % (mail_domain,),
    )
    print("Welcome message sent.")
    setupplugin.message_sent.wait()
    if admingrpid_old is not None:
        setupplugin.message_sent.clear()
        try:
            oldgroup = ac.get_chat_by_id(int(admingrpid_old))
            oldgroup.send_text(
                "Someone created a new admin group on the command line. "
                "This one is not valid anymore.",
            )
            setupplugin.message_sent.wait()
        except ValueError as e:
            print("Could not notify the old admin group:", str(e))
        print("The old admin group was deactivated.")
    sys.stdout.flush()  # flush stdout to actually show the messages above
    ac.shutdown()

    with mailadmdb.write_transaction() as wconn:
        wconn.set_config("admingrpid", chat.id)


@click.command()
@click.pass_context
def config(ctx):
    """show and manipulate config settings."""
    db = get_mailadm_db(ctx)
    with db.read_connection() as conn:
        click.secho("** mailadm version: {}".format(mailadm.__version__))
        click.secho("** mailadm database path: {}".format(db.path))
        for name, val in conn.get_config_items():
            click.secho("{:22s} {}".format(name, val))


@click.command()
@click.pass_context
def list_tokens(ctx):
    """list available tokens"""
    db = get_mailadm_db(ctx)
    click.secho(mailadm.commands.list_tokens(db))


@click.command()
@click.option("--token", type=str, default=None, help="name of token")
@click.pass_context
def list_users(ctx, token):
    """list users"""
    db = get_mailadm_db(ctx)
    with db.read_connection() as conn:
        for user_info in conn.get_user_list(token=token):
            click.secho("{} [{}]".format(user_info.addr, user_info.token_name))


def dump_token_info(token_info):
    click.echo(style("token:{}".format(token_info.name), fg="green"))
    click.echo("  prefix = {}".format(token_info.prefix))
    click.echo("  expiry = {}".format(token_info.expiry))
    click.echo("  maxuse = {}".format(token_info.maxuse))
    click.echo("  usecount = {}".format(token_info.usecount))
    click.echo("  token  = {}".format(token_info.token))
    click.echo("  " + token_info.get_web_url())
    click.echo("  " + token_info.get_qr_uri())


@click.command()
@click.argument("name", type=str, required=True)
@click.option("--expiry", type=str, default="1d", help="account expiry eg 1w 3d -- default is 1d")
@click.option(
    "--maxuse",
    type=int,
    default=50,
    help="maximum number of accounts this token can create",
)
@click.option(
    "--prefix",
    type=str,
    default="tmp.",
    help="prefix for all e-mail addresses for this token",
)
@click.option("--token", type=str, default=None, help="name of token to be used")
@click.pass_context
def add_token(ctx, name, expiry, maxuse, prefix, token):
    """add new token for generating new e-mail addresses"""
    db = get_mailadm_db(ctx)
    result = mailadm.commands.add_token(db, name, expiry, maxuse, prefix, token)
    if result["status"] == "error":
        ctx.fail(result["message"])
    click.secho(result["message"])


@click.command()
@click.argument("name", type=str, required=True)
@click.option(
    "--expiry",
    type=str,
    default=None,
    help="account expiry eg 1w 3d -- default is to not change",
)
@click.option(
    "--maxuse",
    type=int,
    default=None,
    help="maximum number of accounts this token can create, default is not to change",
)
@click.option(
    "--prefix",
    type=str,
    default=None,
    help="prefix for all e-mail addresses for this token, default is not to change",
)
@click.pass_context
def mod_token(ctx, name, expiry, prefix, maxuse):
    """modify a token selectively"""
    db = get_mailadm_db(ctx)

    with db.write_transaction() as conn:
        conn.mod_token(name=name, expiry=expiry, maxuse=maxuse, prefix=prefix)
        tc = conn.get_tokeninfo_by_name(name)
        dump_token_info(tc)


@click.command()
@click.argument("name", type=str, required=True)
@click.pass_context
def del_token(ctx, name):
    """remove named token"""
    db = get_mailadm_db(ctx)
    with db.write_transaction() as conn:
        conn.del_token(name=name)
    click.secho("deleted token " + name)


@click.command()
@click.argument("tokenname", type=str, required=True)
@click.pass_context
def gen_qr(ctx, tokenname):
    """generate qr code image for a token."""
    db = get_mailadm_db(ctx)
    result = mailadm.commands.qr_from_token(db, tokenname)
    if result["status"] == "error":
        ctx.fail(result["message"])
    fn = result["filename"]

    click.secho("{} written for token '{}'".format(fn, tokenname))


@click.command()
@click.option(
    "--web-endpoint",
    type=str,
    help="external URL for Web API create-account requests",
    envvar="WEB_ENDPOINT",
    default="https://example.org/new_email",
    show_default=True,
)
@click.option(
    "--mail-domain",
    type=str,
    help="mail domain for which we create new users",
    envvar="MAIL_DOMAIN",
    default="example.org",
    show_default=True,
)
@click.option(
    "--mailcow-endpoint",
    type=str,
    required=True,
    envvar="MAILCOW_ENDPOINT",
    help="the API endpoint of the mailcow instance",
)
@click.option(
    "--mailcow-token",
    type=str,
    required=True,
    envvar="MAILCOW_TOKEN",
    help="you can get an API token in the mailcow web interface",
)
@click.pass_context
def init(ctx, web_endpoint, mail_domain, mailcow_endpoint, mailcow_token):
    """(re-)initialize configuration in mailadm database.

    Warnings: init can be called multiple times but if you are doing this to a
    database that already has users and tokens, you might run into trouble,
    depending on what you changed.
    """
    db = get_mailadm_db(ctx, fail_missing_config=False)
    click.secho("initializing database {}".format(db.path))

    db.init_config(
        mail_domain=mail_domain,
        web_endpoint=web_endpoint,
        mailcow_endpoint=mailcow_endpoint,
        mailcow_token=mailcow_token,
    )


@click.command()
@click.argument("addr", type=str, required=True)
@click.option(
    "--password",
    type=str,
    default=None,
    help="if not specified, generate a random password",
)
@click.option(
    "--token",
    type=str,
    default=None,
    help="name of token. if not specified, automatically use first token matching addr",
)
@option_dryrun
@click.pass_context
def add_user(ctx, addr, password, token, dryrun):
    """add user as a mailadm managed account."""
    db = get_mailadm_db(ctx)
    result = mailadm.commands.add_user(db, token, addr, password, dryrun)
    if result["status"] == "error":
        ctx.fail(result["message"])
    elif result["status"] == "success":
        click.secho(
            "Created %s with password: %s" % (result["message"].addr, result["message"].password),
        )
    elif result["status"] == "dryrun":
        click.secho(
            "Would create %s with password %s"
            % (result["message"].addr, result["message"].password),
        )


@click.command()
@click.argument("addr", type=str, required=True)
@click.pass_context
def del_user(ctx, addr):
    """remove e-mail address"""
    with get_mailadm_db(ctx).write_transaction() as conn:
        try:
            conn.delete_email_account(addr)
            click.secho("deleted user " + addr)
        except (DBError, MailcowError) as e:
            ctx.fail("failed to delete e-mail account {}: {}".format(addr, e))


@click.command()
@option_dryrun
@click.pass_context
def prune(ctx, dryrun):
    """prune expired users from postfix and dovecot configurations"""
    result = mailadm.commands.prune(get_mailadm_db(ctx), dryrun=dryrun)
    for msg in result.get("message"):
        if result.get("status") == "error":
            ctx.fail(msg)
        else:
            click.secho(msg)


@click.command()
@click.pass_context
@click.option(
    "--debug",
    is_flag=True,
    default=False,
    help="run server in debug mode and don't change any files",
)
def web(ctx, debug):
    """(debugging-only!) serve http account creation Web API on localhost"""
    from .web import create_app_from_db

    db = get_mailadm_db(ctx)
    app = create_app_from_db(db)
    app.run(debug=debug, host="localhost", port=3691)


@click.command()
@click.pass_context
def migrate_db(ctx):
    db = get_mailadm_db(ctx)
    with db.write_transaction() as conn:
        conn.execute("PRAGMA foreign_keys=on;")

        q = "SELECT addr, date, ttl, token_name from users"
        users = [UserInfo(*args) for args in conn.execute(q).fetchall()]
        q = "DROP TABLE users"
        conn.execute(q)

        conn.execute(
            """
                    CREATE TABLE users (
                        addr TEXT PRIMARY KEY,
                        date INTEGER,
                        ttl INTEGER,
                        token_name TEXT NOT NULL,
                        FOREIGN KEY (token_name) REFERENCES tokens (name)
                    )
                """,
        )
        for u in users:
            q = """INSERT INTO users (addr, date, ttl, token_name)
                           VALUES (?, ?, ?, ?)"""
            conn.execute(q, (u.addr, u.date, u.ttl, u.token_name))

        q = "DELETE FROM config WHERE name=?"
        conn.execute(q, ("vmail_user",))
        conn.execute(q, ("path_virtual_mailboxes",))


mailadm_main.add_command(setup_bot)
mailadm_main.add_command(init)
mailadm_main.add_command(config)
mailadm_main.add_command(list_tokens)
mailadm_main.add_command(add_token)
mailadm_main.add_command(mod_token)
mailadm_main.add_command(del_token)
mailadm_main.add_command(gen_qr)
mailadm_main.add_command(add_user)
mailadm_main.add_command(del_user)
mailadm_main.add_command(list_users)
mailadm_main.add_command(prune)
mailadm_main.add_command(web)
mailadm_main.add_command(migrate_db)


if __name__ == "__main__":
    mailadm_main()



import datetime
import os
import time
from random import randint

import pytest
from mailadm.mailcow import MailcowError


@pytest.fixture
def mycmd(cmd, make_db, tmpdir, monkeypatch, mailcow_domain, mailcow_endpoint):
    db = make_db(tmpdir.mkdir("mycmd"), init=False)
    monkeypatch.setenv("MAILADM_DB", str(db.path))
    monkeypatch.setenv("ADMBOT_DB", str(tmpdir.mkdir("admbot")) + "admbot.db")
    cmd.db = db
    if not os.environ["MAILCOW_TOKEN"]:
        raise KeyError("Please set mailcow API Key with the environment variable MAILCOW_TOKEN")
    cmd.run_ok(
        [
            "init",
            "--mailcow-endpoint",
            mailcow_endpoint,
            "--mail-domain",
            mailcow_domain,
            "--web-endpoint",
            "https://example.org/new_email",
        ],
    )
    return cmd


def test_bare(cmd):
    cmd.run_ok(
        [],
        """
        *account creation*
    """,
    )


class TestInitAndInstall:
    def test_init(self, cmd, monkeypatch, tmpdir):
        monkeypatch.setenv("MAILADM_DB", tmpdir.join("mailadm.db").strpath)
        cmd.run_ok(
            [
                "init",
                "--mailcow-endpoint",
                "unfortunately-required",
                "--mailcow-token",
                "unfortunately-required",
            ],
        )


class TestConfig:
    def test_config_simple(self, mycmd):
        mycmd.run_ok(
            ["config"],
            """
            dbversion*
        """,
        )


class TestQR:
    def test_gen_qr(self, mycmd, tmpdir, monkeypatch, mailcow_domain):
        mycmd.run_ok(
            ["add-token", "oneweek", "--token=1w_Zeeg1RSOK4e3Nh0V", "--prefix", "", "--expiry=1w"],
        )
        mycmd.run_ok(
            ["list-tokens"],
            """
            *oneweek*
        """,
        )
        monkeypatch.chdir(tmpdir)
        os.system("mkdir docker-data")
        mycmd.run_ok(
            ["gen-qr", "oneweek"],
            """
            *dcaccount-*-oneweek.png*
        """,
        )
        p = tmpdir.join("docker-data/dcaccount-%s-oneweek.png" % (mailcow_domain,))
        assert p.exists()

    def test_gen_qr_no_token(self, mycmd):
        mycmd.run_fail(
            ["gen-qr", "notexistingtoken"],
            """
            *Error*not*
        """,
        )


class TestTokens:
    def test_uninitialized(self, cmd):
        cmd.run_fail(
            ["list-tokens"],
            """
            *MAILADM_DB not set*
        """,
        )

    def test_tokens(self, mycmd):
        mycmd.run_ok(
            ["add-token", "oneweek", "--token=1w_Zeeg1RSOK4e3Nh0V", "--prefix", "", "--expiry=1w"],
        )
        mycmd.run_ok(
            ["list-tokens"],
            """
            *oneweek*
            *https://example.org*
            *DCACCOUNT*
        """,
        )

    @pytest.mark.parametrize("i", range(3))
    def test_tokens_add(self, mycmd, i):
        mycmd.run_ok(
            ["add-token", "test1", "--expiry=1d", "--prefix=tmpy."],
            """
            *DCACCOUNT*&n=test1
        """,
        )
        out = mycmd.run_ok(
            ["list-tokens"],
            """
            *of 50 times*
            *DCACCOUNT*&n=test1
        """,
        )
        for line in out.splitlines():
            parts = line.split(":")
            if len(parts) >= 2 and parts[0].strip() == "token":
                token = parts[1].strip().replace("_", "")
                assert token.isalnum()
                break
        else:
            pytest.fail()

        mycmd.run_ok(
            ["del-token", "test1"],
            """
            *deleted*test1*
        """,
        )
        out = mycmd.run_ok(["list-tokens"])
        assert "test1" not in out

    def test_tokens_add_maxuse(self, mycmd):
        mycmd.run_ok(
            ["add-token", "test1", "--maxuse=10"],
            """
            *of 10 times*
            *DCACCOUNT*&n=test1
        """,
        )
        mycmd.run_ok(
            ["list-tokens"],
            """
            *of 10 times*
            *DCACCOUNT*&n=test1
        """,
        )
        mycmd.run_ok(["mod-token", "--maxuse=1000", "test1"])
        mycmd.run_ok(
            ["list-tokens"],
            """
            *of 1000 times*
            *DCACCOUNT*&n=test1
        """,
        )


class TestUsers:
    def test_adduser_help(self, mycmd):
        mycmd.run_ok(
            ["add-user", "-h"],
            """
            *add*user*
        """,
        )

    def test_add_del_user(self, mycmd, mailcow_domain):
        mycmd.run_ok(["add-token", "test1", "--expiry=1d", "--prefix", "pytest."])
        addr = "pytest.%s@%s" % (randint(0, 99999), mailcow_domain)
        mycmd.run_ok(
            ["add-user", addr],
            """
            *Created*pytest*@*
        """,
        )
        mycmd.run_ok(
            ["list-users"],
            """
            *{addr}*
        """.format(
                addr=addr,
            ),
        )
        mycmd.run_fail(
            ["add-user", addr],
            """
            *failed to add*pytest* account does already exist*
        """,
        )
        mycmd.run_ok(
            ["del-user", addr],
            """
            *deleted*pytest*@*
        """,
        )
        mycmd.run_ok(
            ["add-user", addr, "--dryrun"],
            """
            *Would create pytest*@*
        """,
        )
        mycmd.run_fail(
            ["del-user", addr],
            """
            *failed to delete*pytest*@*does not exist*
        """,
        )

    def test_adduser_and_expire(self, mycmd, monkeypatch, mailcow_domain):
        mycmd.run_ok(["add-token", "test1", "--expiry=1d", "--prefix", "pytest."])
        addr = "pytest.%s@%s" % (randint(0, 49999), mailcow_domain)
        mycmd.run_ok(
            ["add-user", addr],
            """
            *Created*pytest*@*
        """,
        )

        to_expire = time.time() - datetime.timedelta(weeks=1).total_seconds() - 1

        # create an old account that should expire
        with monkeypatch.context() as m:
            m.setattr(time, "time", lambda: to_expire)
            addr2 = "pytest.%s@%s" % (randint(50000, 99999), mailcow_domain)
            mycmd.run_ok(
                ["add-user", addr2],
                """
                *Created*pytest*@*
            """,
            )

        out = mycmd.run_ok(["list-users"])
        assert addr2 in out

        mycmd.run_ok(["prune"])

        out = mycmd.run_ok(["list-users"])
        assert addr in out
        assert addr2 not in out

        mycmd.run_ok(["del-user", addr])

    def test_two_tokens_users(self, mycmd, mailcow_domain):
        mycmd.run_ok(["add-token", "test1", "--expiry=1d", "--prefix=tmpy."])
        mycmd.run_ok(["add-token", "test2", "--expiry=1d", "--prefix=tmpx."])
        mycmd.run_fail(["add-user", "x@" + mailcow_domain])
        addr = "tmpy.%s@%s" % (randint(0, 49999), mailcow_domain)
        addr2 = "tmpx.%s@%s" % (randint(50000, 99999), mailcow_domain)
        mycmd.run_ok(["add-user", addr])
        mycmd.run_ok(["add-user", addr2])
        mycmd.run_ok(
            ["list-users"],
            """
            tmpy.*test1*
            tmpx.*test2*
        """,
        )
        out = mycmd.run_ok(["list-users", "--token", "test1"])
        assert addr in out
        assert addr2 not in out
        out = mycmd.run_ok(["list-users", "--token", "test2"])
        assert addr not in out
        assert addr2 in out
        mycmd.run_ok(["del-user", addr])
        mycmd.run_ok(["del-user", addr2])


class TestSetupBot:
    def test_account_already_exists(self, mycmd, mailcow, mailcow_domain):
        print(os.environ.items())
        delete_later = True
        try:
            mailcow.add_user_mailcow("mailadm@" + mailcow_domain, "asdf1234", "pytest")
        except MailcowError as e:
            if "object_exists" in str(e):
                delete_later = False
        mycmd.run_fail(
            ["setup-bot"],
            """
            *mailadm@* already exists; delete the account in mailcow or specify*
        """,
        )
        if delete_later:
            mailcow.del_user_mailcow("mailadm@" + mailcow_domain)

    def test_specify_addr_not_password(self, mycmd):
        mycmd.run_fail(
            ["setup-bot", "--email", "bot@example.org"],
            """
            *You need to provide --password if you want to use an existing account*
        """,
        )

    def test_specify_password_not_addr(self, mycmd):
        mycmd.run_fail(
            ["setup-bot", "--password", "asdf"],
            """
            *Please also provide --email to use an email account for the mailadm*
        """,
        )

    def test_wrong_credentials(self, mycmd):
        mycmd.run_fail(
            ["setup-bot", "--email", "bot@testrun.org", "--password", "asdf"],
            """
            *Cannot login as "bot@testrun.org". Please check if the email address and the password*
        """,
        )



import sys

import pytest
from mailadm.util import get_human_readable_id, parse_expiry_code


@pytest.mark.parametrize(
    ("code", "duration"),
    [
        ("never", sys.maxsize),
        ("2y", 2 * 365 * 24 * 60 * 60),
        ("1w", 7 * 24 * 60 * 60),
        ("2w", 2 * 7 * 24 * 60 * 60),
        ("2d", 2 * 24 * 60 * 60),
        ("5h", 5 * 60 * 60),
        ("15h", 15 * 60 * 60),
        ("20s", 20),
        ("0h", 0),
    ],
)
def test_parse_expiries(code, duration):
    res = parse_expiry_code(code)
    assert res == duration


def test_parse_expiries_short():
    with pytest.raises(ValueError):
        parse_expiry_code("h")


def test_parse_expiries_wrong():
    with pytest.raises(ValueError):
        parse_expiry_code("123h123d")
    with pytest.raises(ValueError):
        parse_expiry_code("12j")


def test_human_readable_id():
    s = get_human_readable_id(len=20)
    assert s.isalnum()



from mailadm.gen_qr import gen_qr
from pyzbar.pyzbar import decode


def test_gen_qr(db):
    with db.write_transaction() as conn:
        config = conn.config
        token = conn.add_token("burner1", expiry="1w", token="1w_7wDioPeeXyZx96v3", prefix="pp")

    image = gen_qr(config=config, token_info=token)
    qr_decoded = decode(image)[0]
    assert bytes(token.get_qr_uri(), encoding="ascii") == qr_decoded.data



import time

import deltachat
import pytest
from deltachat.capi import lib as dclib

TIMEOUT = 45


@pytest.mark.timeout(TIMEOUT)
class TestAdminGroup:
    def test_help(self, admingroup):
        admingroup.send_text("/help")
        while "/add-user" not in admingroup.get_messages()[len(admingroup.get_messages()) - 1].text:
            print(admingroup.get_messages()[len(admingroup.get_messages()) - 1].text)
            time.sleep(1)
        reply = admingroup.get_messages()[len(admingroup.get_messages()) - 1]
        assert reply.text.startswith("/add-user addr password token")

    def test_list_tokens(self, admingroup):
        command = admingroup.send_text("/list-tokens")
        while "Existing" not in admingroup.get_messages()[len(admingroup.get_messages()) - 1].text:
            print(admingroup.get_messages()[len(admingroup.get_messages()) - 1].text)
            time.sleep(1)
        reply = admingroup.get_messages()[len(admingroup.get_messages()) - 1]
        assert reply.text.startswith("Existing tokens:")
        assert reply.quote == command

    def test_wrong_number_of_arguments(self, admingroup):
        command = admingroup.send_text("/add-token pytest")
        while "Sorry" not in admingroup.get_messages()[len(admingroup.get_messages()) - 1].text:
            print(admingroup.get_messages()[len(admingroup.get_messages()) - 1].text)
            time.sleep(1)
        reply = admingroup.get_messages()[len(admingroup.get_messages()) - 1]
        assert reply.quote == command
        print(reply.text)

    @pytest.mark.skip("This test works in real life, but not under test conditions somehow")
    def test_check_privileges(self, admingroup):
        direct = admingroup.botadmin.create_chat(admingroup.admbot.get_config("addr"))
        direct.send_text("/list-tokens")
        while "Sorry, I" not in direct.get_messages()[-1].text:
            print(direct.get_messages()[-1].text)
            time.sleep(0.1)
        assert direct.get_messages()[-1].text == "Sorry, I only take commands from the admin group."


@pytest.mark.timeout(TIMEOUT * 2)
class TestSupportGroup:
    def test_support_group_relaying(self, admingroup, supportuser):
        class SupportGroupUserPlugin:
            def __init__(self, account, supportuser):
                self.account = account
                self.account.add_account_plugin(deltachat.events.FFIEventLogger(self.account))
                self.supportuser = supportuser

            @deltachat.account_hookimpl
            def ac_incoming_message(self, message: deltachat.Message):
                message.create_chat()

                assert len(message.chat.get_contacts()) == 2
                assert message.override_sender_name == self.supportuser.get_config("addr")

                if message.text == "Can I ask you a support question?":
                    message.chat.send_text("I hope the user can't read this")
                    print("\n  botadmin to supportgroup: I hope the user can't read this\n")
                    reply = deltachat.Message.new_empty(self.account, "text")
                    reply.set_text("Yes of course you can ask us :)")
                    reply.quote = message
                    message.chat.send_msg(reply)
                    print("\n  botadmin to supportgroup: Yes of course you can ask us :)\n")
                else:
                    print("\n  botadmin received:", message.text, "\n")

        supportchat = supportuser.create_chat(admingroup.admbot.get_config("addr"))
        question = "Can I ask you a support question?"
        supportchat.send_text(question)
        print("\n  supportuser to supportchat: Can I ask you a support question?\n")
        admin = admingroup.botadmin
        admin.add_account_plugin(SupportGroupUserPlugin(admin, supportuser))
        while len(admin.get_chats()) < 2:
            time.sleep(0.1)
        # AcceptChatPlugin will send 2 messages to the support group now
        support_group_name = supportuser.get_config("addr") + " support group"
        for chat in admin.get_chats():
            print(chat.get_name() + str(chat.id))
        supportgroup = next(
            filter(lambda chat: chat.get_name() == support_group_name, admin.get_chats()),
        )
        while "Yes of" not in supportchat.get_messages()[len(supportchat.get_messages()) - 1].text:
            print(supportchat.get_messages()[len(supportchat.get_messages()) - 1].text)
            time.sleep(1)
        botreply = supportchat.get_messages()[1]
        assert botreply.text == "Yes of course you can ask us :)"
        supportchat.send_text("Okay, I will think of something :)")
        print("\n  supportuser to supportchat: Okay, I will think of something :)\n")
        while "Okay," not in supportgroup.get_messages()[len(supportgroup.get_messages()) - 1].text:
            print(supportchat.get_messages()[len(supportchat.get_messages()) - 1].text)
            time.sleep(1)
        assert "I hope the user can't read this" not in [
            msg.text for msg in supportchat.get_messages()
        ]

    def test_invite_bot_to_group(self, admingroup, supportuser):
        botcontact = supportuser.create_contact(admingroup.admbot.get_config("addr"))
        false_group = supportuser.create_group_chat("invite bot", [botcontact])
        false_group.send_text("Welcome, bot!")
        while "left by %s" % (botcontact.addr,) not in false_group.get_messages()[-1].text:
            print(false_group.get_messages()[-1].text)
            time.sleep(0.1)
        assert len(false_group.get_contacts()) == 1
        sorry_message = "Sorry, you can not contact me in groups. Please use a 1:1 chat."
        assert false_group.get_messages()[-2].text == sorry_message

    def test_bot_receives_system_message(self, admingroup):
        def get_group_chats(account):
            group_chats = []
            for chat in ac.get_chats():
                if chat.is_group():
                    group_chats.append(chat)
            return group_chats

        ac = admingroup.admbot
        num_chats = len(get_group_chats(ac))
        # put system message in admbot's INBOX
        dev_msg = deltachat.Message.new_empty(ac, "text")
        dev_msg.set_text("This shouldn't create a support group")
        dclib.dc_add_device_msg(ac._dc_context, bytes("test_device_msg", "ascii"), dev_msg._dc_msg)
        # assert that admbot didn't create a support group
        assert num_chats == len(get_group_chats(ac))

    def test_did_bot_create_support_group(self, admingroup, supportuser):
        # send first message to support user to test that it isn't seen as support group later
        bot = admingroup.admbot
        supportchat_bot_side = bot.create_chat(supportuser.get_config("addr"))
        supportchat_bot_side.send_text("Your account will expire soon!")

        # create support group
        supportchat = supportuser.create_chat(admingroup.admbot.get_config("addr"))
        question = "Can I ask you a support question?"
        supportchat.send_text(question)
        support_group_name = supportuser.get_config("addr") + " support group"

        # wait for supportgroup to be created
        while 1:
            try:
                supportgroup = next(
                    filter(lambda chat: chat.get_name() == support_group_name, bot.get_chats()),
                )
                print(supportgroup.get_messages()[0].get_sender_contact().addr)
            except (StopIteration, IndexError):
                time.sleep(0.1)
            else:
                break
        print(bot.get_self_contact().addr)

        assert admingroup.botplugin.is_support_group(supportgroup)
        assert not admingroup.botplugin.is_support_group(admingroup)
        assert not admingroup.botplugin.is_support_group(supportchat_bot_side)

    def test_support_user_help(self, admingroup, supportuser):
        supchat = supportuser.create_chat(admingroup.admbot.get_config("addr"))
        supchat.send_text("/help")
        while "use this chat to talk to the admins." not in supchat.get_messages()[-1].text:
            time.sleep(0.1)
        supchat.send_text("/list-tokens")
        while "Sorry, I" not in supchat.get_messages()[-1].text:
            print(supchat.get_messages()[-1].text)
            time.sleep(0.1)
        assert supchat.get_messages()[-1].text == "Sorry, I only take commands in the admin group."



import email
import smtplib
import urllib.parse

import requests
from imapclient import IMAPClient


def receive_imap(host, user, password, num):
    conn = IMAPClient(host)
    conn.login(user, password)
    info = conn.select_folder("INBOX")
    print(info)
    messages = conn.fetch("1:*", [b"FLAGS"])
    latest_msg = max(messages)
    requested = "FLAGS BODY.PEEK[]"
    for uid, data in conn.fetch([latest_msg], [b"FLAGS", b"BODY.PEEK[]"]).items():
        body_bytes = data[b"BODY[]"]
        email_message = email.message_from_bytes(body_bytes)
        assert email_message["subject"] == str(num)
        print("received message num={}".format(num))
        print(email_message)


def send_self_mail(host, user, password, num):
    smtp = smtplib.SMTP(host, 587)
    smtp.set_debuglevel(1)
    smtp.starttls()
    smtp.login(user, password)
    msg = """\
From: {user}
To: {user}
Subject: {num}

hi there
""".format(
        user=user, num=num
    )
    smtp.sendmail(user, [user], msg)
    print("send message num={}".format(num))


if __name__ == "__main__":
    import sys

    if sys.argv[1].startswith("https"):
        res = requests.post(sys.argv[1])
        assert res.status_code == 200
        data = res.json()
        user = data["email"]
        password = data["password"]

        if len(sys.argv) >= 3:
            host = sys.argv[2]
        else:
            host = urllib.parse.urlparse(sys.argv[1]).hostname
    else:
        user, password, host = sys.argv[1:]
    num = 42
    send_self_mail(host, user, password, num)
    receive_imap(host, user, password, num)



from random import randint

import pytest
from mailadm.mailcow import MailcowError


class TestMailcow:
    def test_get_users(self, mailcow):
        mailcow.get_user_list()

    def test_add_del_user(self, mailcow, mailcow_domain):
        addr = "pytest.%s@%s" % (randint(0, 99999), mailcow_domain)
        mailcow.add_user_mailcow(addr, "asdf1234", "pytest")

        user = mailcow.get_user(addr)
        assert user.addr == addr
        assert user.token == "pytest"

        mailcow.del_user_mailcow(addr)
        assert mailcow.get_user(addr) is None

    def test_wrong_token(self, mailcow, mailcow_domain):
        mailcow.auth = {"X-API-Key": "asdf"}
        addr = "pytest.%s@%s" % (randint(0, 99999), mailcow_domain)
        with pytest.raises(MailcowError):
            mailcow.get_user_list()
        with pytest.raises(MailcowError):
            mailcow.add_user_mailcow(addr, "asdf1234", "pytest")
        with pytest.raises(MailcowError):
            mailcow.get_user(addr)
        with pytest.raises(MailcowError):
            mailcow.del_user_mailcow(addr)



import pytest
from mailadm.conn import DBError, TokenExhaustedError, UserNotFoundError
from mailadm.util import gen_password


def test_token(tmpdir, make_db):
    db = make_db(tmpdir)
    with db._get_connection(closing=True, write=True) as conn:
        assert not conn.get_token_list()
        conn.add_token(name="pytest:1w", prefix="xyz", expiry="1w", maxuse=5, token="1234567890123")
        conn.commit()
    with db._get_connection(closing=True) as conn:
        assert len(conn.get_token_list()) == 1
        entry = conn.get_tokeninfo_by_name("pytest:1w")
        assert entry.expiry == "1w"
        assert entry.prefix == "xyz"
        assert entry.usecount == 0

        entry = conn.get_tokeninfo_by_token("1234567890123")
        assert entry.name == "pytest:1w"
        assert entry.expiry == "1w"
        assert entry.prefix == "xyz"
        assert entry.maxuse == 5

    with db.write_transaction() as conn:
        assert conn.get_token_list()
        conn.del_token(name="pytest:1w")
        assert not conn.get_token_list()


class TestTokenAccounts:
    MAXUSE = 10

    @pytest.fixture
    def conn(self, tmpdir, make_db):
        db = make_db(tmpdir.mkdir("conn"))
        conn = db._get_connection(write=True)
        conn.add_token(
            name="pytest:1h",
            prefix="xyz",
            expiry="1h",
            maxuse=self.MAXUSE,
            token="123456789012345",
        )
        conn.commit()
        return conn

    def test_add_with_wrong_token(self, conn, mailcow_domain):
        now = 10000
        addr = "tmp.123@" + mailcow_domain
        with pytest.raises(DBError):
            conn.add_user_db(addr=addr, date=now, ttl=60 * 60, token_name="112l3kj123123")

    def test_add_maxuse(self, conn, mailcow_domain):
        now = 10000
        password = gen_password()
        for i in range(self.MAXUSE):
            addr = "tmp.%s@%s" % (i, mailcow_domain)
            conn.add_user_db(addr=addr, date=now, ttl=60 * 60, token_name="pytest:1h")

        token_info = conn.get_tokeninfo_by_name("pytest:1h")
        with pytest.raises(TokenExhaustedError):
            conn.add_email_account(token_info, addr="tmp.xx@" + mailcow_domain, password=password)

    def test_add_expire_del(self, conn, mailcow_domain):
        now = 10000
        addr = "tmp.123@" + mailcow_domain
        addr2 = "tmp.456@" + mailcow_domain
        addr3 = "tmp.789@" + mailcow_domain
        conn.add_user_db(addr=addr, date=now, ttl=60 * 60, token_name="pytest:1h")
        with pytest.raises(DBError):
            conn.add_user_db(addr=addr, date=now, ttl=60 * 60, token_name="pytest:1h")
        conn.add_user_db(addr=addr2, date=now, ttl=30 * 60, token_name="pytest:1h")
        conn.add_user_db(addr=addr3, date=now, ttl=32 * 60, token_name="pytest:1h")
        conn.commit()
        expired = conn.get_expired_users(sysdate=now + 31 * 60)
        assert len(expired) == 1
        assert expired[0].addr == addr2

        users = conn.get_user_list(token="pytest:1h")
        assert len(users) == 3
        conn.del_user_db(addr2)
        conn.commit()
        assert len(conn.get_user_list(token="pytest:1h")) == 2
        addrs = [u.addr for u in conn.get_user_list(token="pytest:1h")]
        assert addrs == [addr, addr3]
        with pytest.raises(UserNotFoundError):
            conn.del_user_db(addr2)
        assert conn.get_tokeninfo_by_name("pytest:1h").usecount == 3
